/* roadmap_category.c - Manage Point of interest categories.
 *
 * LICENSE:
 *
 * Copyright 2002 Pascal F. Martin
 *
 * This file is part of RoadMap.
 *
 * RoadMap is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * RoadMap is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with RoadMap; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESCRIPTION:
 *
 * This module manages a tree of point of interest categories. The full
 * description of each category is managed separately from the points
 * themselves, except that each point database reference each category
 * using its unique (internal) name.
 *
 * The categories are stored in a sqlite database, so that new categories
 * can be added by the user when needed. In this respect, there is no
 * "map" of categories provided with RoadMap.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "roadmap.h"
#include "roadmap_category.h"

#include <sqlite.h>

static const char *RoadMapCategoryLocalName = "rdmcat.sld";

static const char *RoadMapCategoryQueryIndex =
    "select ROWID, NAME from ROADMAP_CATEGORIES order by ROWID";

static sqlite *RoadMapCategoryLocal = NULL;


#define ROADMAP_CATEGORY_BLOCK   128

struct roadmap_category_index {
   int   rowid;
   char *name;
}

static struct roadmap_category_index *RoadMapCategoryIndex = NULL;
static int RoadMapCategoryCount = 0;
static int RoadMapCategorySpace = 0;

static int roadmap_category_make_index (void *context,
                                        int argc,
                                        char **argv, char **columns) {

   char *name;
   int   rowid;

   if (argc != 2) return 1;

   if (strcmp (columns[0], "ROWID") == 0)
   {
      rowid = atoi(argv[0]);
      name  = argv[1];
   }
   else
   {
      rowid = atoi(argv[1]);
      name  = argv[0];
   }

   if (RoadMapCategoryCount >= RoadMapCategorySpace) {

      RoadMapCategorySpace += ROADMAP_CATEGORY_BLOCK;
      RoadMapCategoryIndex =
         realloc(RoadMapCategoryIndex, RoadMapCategorySpace);
      roadmap_check_allocated(RoadMapCategoryIndex);
   }

   return 0;
}

void roadmap_category_initialize (void) {

   char *tail;
   char *error;
   int   status;
   sqlite_vm *cursor;

   for (path = roadmap_path_first();
        path != NULL;
        path = roadmap_path_next(path)) {

      if (roadmap_file_exists (path, RoadMapCategoryLocalName)) {

         RoadMapCategoryLocal =
            sqlite_open
               (roadmap_path_join (path, RoadMapCategoryLocalName),
                0,
                &error);

         if (RoadMapCategoryLocal == NULL) {
            roadmap_log (ROADMAP_ERROR,
                         "cannot open category database %s/%s",
                         path, RoadMapCategoryLocalName);
            free(error);
            return;
         }

         status = sqlite_exec (RoadMapCategoryLocal,
                               RoadMapCategoryQueryIndex,
                               roadmap_category_make_index,
                               NULL,
                               &error);

         if (status != SQLITE_OK) {
            roadmap_log (ROADMAP_ERROR,
                         "error %d when indexing database %s/%s",
                         path,
                         RoadMapCategoryLocalName);
            free(error);
            sqlite_close (RoadMapCategoryLocal);
            RoadMapCategoryLocal = NULL;
            return;
         }

         break;
      }
   }
}

int roadmap_category_get_id (const char *name) {

   if (RoadMapCategoryLocal == NULL) return -1;
   if (sqlite_exec(RoadMapCategoryLocal, "select
}

const char *roadmap_category_get_description (int id) {

   if (RoadMapCategoryLocal == NULL) return NULL;
}

const char *roadmap_category_get_icon (int id) {

   if (RoadMapCategoryLocal == NULL) return NULL;
}

int  roadmap_category_get_first (int parent_id) {

   if (RoadMapCategoryLocal == NULL) return -1;
}

int  roadmap_category_get_next (int sibling) {

   if (RoadMapCategoryLocal == NULL) return -1;
}

void roadmap_category_hide (const char *name, int recursive) {

   if (RoadMapCategoryLocal == NULL) return;
}

void roadmap_category_show (const char *name) {

   if (RoadMapCategoryLocal == NULL) return;
}

int  roadmap_category_is_visible (const char *name) {

   if (RoadMapCategoryLocal == NULL) return 0;
}

int  roadmap_category_add (const char *parent,
                           const char *name, const char *description) {
}

