/* roadmap_category.h - Manage Point of interest categories.
 *
 * LICENSE:
 *
 * Copyright 2002 Pascal F. Martin
 *
 * This file is part of RoadMap.
 *
 * RoadMap is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * RoadMap is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with RoadMap; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESCRIPTION:
 *
 * This module manages a tree of point of interest categories. The full
 * description of each category is managed separately from the points
 * themselves, except that each point database reference each category
 * using its unique (internal) name.
 *
 * Why is it done this way? there are two constraints that RoadMap must
 * deal with:
 * - RoadMap must know for once the list of categories in order to
 *   identify them using a "stable" numerical id (which id is created
 *   on the fly by RoadMap, not really saved on disk..).
 * - each point database should not repeat all category information, as
 *   this could lead to inconsistent category information. In addition,
 *   this also saves space (the unique name can be short).
 *
 * A category is identified by a unique name to minimize the risk of
 * conflicts between databases of different origins. By convention,
 * the category named "ROOT" is the parent of all top-level categories.
 *
 * Another convention is that the name representing a public company
 * (for example Mc Donald's) should be its stock ticker prefixed with
 * "ST:" (for example "ST:MCD" to represent Mc Donald's).
 *
 * Other standard categories are listed in the category reference file
 * (see category.csv).
 */

#ifndef INCLUDE__ROADMAP_CATEGORY__H
#define INCLUDE__ROADMAP_CATEGORY__H

#include "roadmap_types.h"

void roadmap_category_initialize (void);

int         roadmap_category_get_id          (const char *name);
const char *roadmap_category_get_description (int id);
const char *roadmap_category_get_icon        (int id);

int  roadmap_category_get_first (int parent_id);
int  roadmap_category_get_next  (int sibling);

void roadmap_category_hide   (const char *name, int recursive);
void roadmap_category_show   (const char *name);

int  roadmap_category_is_visible (const char *name);

int  roadmap_category_add (const char *parent,
                           const char *name, const char *description);

#endif // INCLUDE__ROADMAP_CATEGORY__H
